"""
   lookupy.tests
   ~~~~~~~~~~~~~

   This module contains tests for the lookupy module written using
   nose to be run using::

       $ nosetests -v

"""

import re
from nose.tools import assert_list_equal, assert_equal, assert_raises

from .lookupy import filter_items, lookup, include_keys, Q, QuerySet, \
    Collection, LookupyError
from .dunderkey import dunderkey, dunder_partition, dunder_init, dunder_last, \
    dunder_get, undunder_keys, dunder_truncate


entries_fixtures = [{'request': {'url': 'http://example.com', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive'}]},
                     'response': {'status': 404, 'headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'},
                                                             {'name': 'Content-Type', 'value': 'text/html'}]}},
                    {'request': {'url': 'http://example.org', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive'}]},
                     'response': {'status': 200, 'headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'},
                                                             {'name': 'Content-Type', 'value': 'text/html'}]}},
                    {'request': {'url': 'http://example.com/myphoto.jpg', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive'}]},
                     'response': {'status': 200, 'headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'},
                                                             {'name': 'Content-Type', 'value': 'image/jpg'}]}}]


def fe(entries, *args, **kwargs):
    return list(filter_items(entries, *args, **kwargs))


def ik(entries, fields):
    return list(include_keys(entries, fields))


## Tests




def test_Collection():
    c = Collection(entries_fixtures)
    assert_list_equal(list(c), entries_fixtures)
    assert_list_equal(list(c), entries_fixtures)


def test_Q():
    entries = entries_fixtures
    q1 = Q(response__status__exact=404, request__url__contains='.com')
    assert q1.evaluate(entries[0])

    # test with negation
    q2 = ~Q(response__status__exact=404)
    assert q2.evaluate(entries[1])
    # test multiple application of negation
    assert not (~q2).evaluate(entries[1])

    q3 = Q(response__status=200)
    assert not (q1 & q3).evaluate(entries[0])
    assert (q1 | q3).evaluate(entries[0])
    assert (~(q1 & q3)).evaluate(entries[0])

    assert_list_equal(list(((Q(request__url__endswith='.jpg') | Q(response__status=404)).evaluate(e)
                      for e in entries)),
                      [True, False, True])

    assert_list_equal(list(((~Q(request__url__endswith='.jpg') | Q(response__status=404)).evaluate(e)
                      for e in entries)),
                      [True, True, False])


def test_lookup():
    entry1, entry2, entry3 = entries_fixtures
    # exact       -- works for strings and int
    assert lookup('request__url__exact', 'http://example.com', entry1)
    assert not lookup('request_url__exact', 'http://example.org', entry1)
    assert lookup('response__status__exact', 404, entry1)
    assert not lookup('response__status__exact', 404, entry2)
    assert lookup('response_unknown__exact', None, entry1)

    # neq         -- works for strings and ints
    assert not lookup('request__url__neq', 'http://example.com', entry1)
    assert lookup('request_url__neq', 'http://example.org', entry1)
    assert not lookup('response__status__neq', 404, entry1)
    assert lookup('response__status__neq', 404, entry2)
    assert not lookup('response_unknown__neq', None, entry1)

    # contains    -- works for strings, else raises error
    assert lookup('request__url__contains', '.com', entry1)
    assert not lookup('request__url__contains', 'www', entry1)
    assert_raises(LookupyError, lookup, 'response__status__contains',
                  2, entry2)
    assert_raises(LookupyError, lookup, 'response__unknown__contains',
                  None, entry2)

    # icontains   -- works for strings, else raises error
    assert lookup('request__url__icontains', 'EXAMPLE', entry1)
    assert not lookup('request__url__icontains', 'www', entry1)
    assert_raises(LookupyError, lookup, 'response__status__icontains',
                  2, entry2)
    assert_raises(LookupyError, lookup,
                  'response__unknown__icontains', None, entry2)

    # in          -- works for strings and lists, else raises error
    assert lookup('request__url__in', ['http://example.com',
                                       'http://blog.example.com'], entry1)

    assert lookup('response__status__in', [400, 200], entry2)
    assert not lookup('response__status__in', [], entry2)
    assert lookup('request__url__in', 'http://example.com/?q=hello', entry1)
    assert_raises(LookupyError, lookup, 'response__status__in', 404, entry1)

    # startswith  -- works for strings, else raises error
    assert lookup('request__url__startswith', 'http://', entry1)
    assert not lookup('request__url__startswith', 'HTTP://', entry1)
    assert_raises(LookupyError, lookup,
                  'response__status__startswith', 4, entry1)

    # istartswith -- works for strings, else raises error
    assert lookup('request__url__istartswith', 'http://', entry1)
    assert lookup('request__url__istartswith', 'HTTP://', entry1)
    assert_raises(LookupyError, lookup,
                  'response__status__istartswith', 4, entry1)

    # endswith    -- works for strings, else raises error
    assert lookup('request__url__endswith', '.jpg', entry3)
    assert not lookup('request__url__endswith', '.JPG', entry3)
    assert_raises(LookupyError, lookup, 'response__status__endswith',
                  0, entry3)

    # iendswith   -- works for strings, else raises error
    assert lookup('request__url__iendswith', '.jpg', entry3)
    assert lookup('request__url__iendswith', '.JPG', entry3)
    assert_raises(LookupyError, lookup, 'response__status__iendswith',
                  0, entry3)

    # gt          -- works for strings and int
    assert lookup('response__status__gt', 200, entry1)
    assert not lookup('response__status__gt', 404, entry1)
    assert lookup('request__url__gt', 'ftp://example.com', entry1)
    assert not lookup('request__url__gt', 'http://example.com', entry1)

    # gte         -- works for strings and int
    assert lookup('response__status__gte', 200, entry1)
    assert lookup('response__status__gte', 404, entry1)
    assert lookup('request__url__gte', 'ftp://example.com', entry1)
    assert lookup('request__url__gte', 'http://example.com', entry1)

    # lt          -- works for strings and int
    assert lookup('response__status__lt', 301, entry2)
    assert not lookup('response__status__lt', 200, entry2)
    assert lookup('request__url__lt', 'ws://example.com', entry2)
    assert not lookup('request__url__lt', 'http://example.org', entry2)

    # lte         -- works for strings and int
    assert lookup('response__status__lte', 301, entry2)
    assert lookup('response__status__lte', 200, entry2)
    assert lookup('request__url__lte', 'ws://example.com', entry2)
    assert lookup('request__url__lte', 'http://example.org', entry2)

    # regex       -- works for compiled patterns and strings
    pattern = r'^http:\/\/.+g$'
    assert lookup('request__url__regex', pattern, entry2)
    assert lookup('request__url__regex', pattern, entry3)
    assert not lookup('request__url__regex', pattern, entry1)
    compiled_pattern = re.compile(pattern)
    assert lookup('request__url__regex', compiled_pattern, entry2)
    assert lookup('request__url__regex', compiled_pattern, entry3)
    assert not lookup('request__url__regex', compiled_pattern, entry1)

    # filter      -- works for Q objects, else raises error
    assert lookup('response__headers__filter',
                  Q(name__exact='Content-Type', value__exact='image/jpg'),
                  entry3)
    assert not lookup('response__headers__filter',
                      Q(name__exact='Content-Type', value__exact='text/html'),
                      entry3)
    assert_raises(LookupyError, lookup, 'response__headers__filter',
                  0, entry3)
    assert_raises(LookupyError, lookup, 'response__headers__filter',
                  "hello", entry3)
    assert_raises(LookupyError, lookup, 'response__headers__filter',
                  None, entry3)
    assert_raises(LookupyError, lookup, 'response__headers__filter',
                  {'a': 'b'}, entry3)
    assert_raises(LookupyError, lookup, 'response__status__filter',
                  Q(name__exact='Content-Type', value__exact='image/jpg'),
                  entry3)

    # nothing     -- works for strings and int
    assert lookup('request__url', 'http://example.com', entry1)
    assert not lookup('request_url', 'http://example.org', entry1)
    assert lookup('response__status', 404, entry1)
    assert not lookup('response__status', 404, entry2)
    assert lookup('response_unknown', None, entry1)


def test_filter_items():
    entries = entries_fixtures

    # when no lookup kwargs passed, all entries are returned
    assert_list_equal(fe(entries), entries)

    # simple 1st level lookups
    assert_list_equal(fe(entries, request__url='http://example.com'), entries[0:1])
    assert_list_equal(fe(entries, response__status=200), entries[1:])
    assert len(fe(entries, response__status=405)) == 0

    # testing compund lookups
    assert len(fe(entries, Q(request__url__exact='http://example.org'))) == 1
    assert len(fe(entries,
                  Q(request__url__exact='http://example.org', response__status=200)
                  |
                  Q(request__url__endswith='.com', response__status=404))) == 2

    assert len(fe(entries,
                  ~Q(request__url__exact='http://example.org', response__status__gte=500)
                  |
                  Q(request__url__endswith='.com', response__status=404))) == 3

    assert len(fe(entries,
                  ~Q(request__url__exact='http://example.org', response__status__gte=500)
                  |
                  Q(request__url__endswith='.com', response__status=404),
                  response__status__exact=200)) == 2


def test_include_keys():
    entries = entries_fixtures
    assert_list_equal(ik(entries, ['request']),
                      [{'request': {'url': 'http://example.com', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive'}]}},
                       {'request': {'url': 'http://example.org', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive'}]}},
                       {'request': {'url': 'http://example.com/myphoto.jpg', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive'}]}}])

    assert_list_equal(ik(entries, ['response__status']),
                      [{'response__status': 404},
                       {'response__status': 200},
                       {'response__status': 200}])

    # when an empty list is passed as fields
    assert_list_equal(ik(entries, []), [{},{},{}])

    # when a non-existent key is passed in fields
    assert_list_equal(ik(entries, ['response__status', 'cookies']),
                      [{'response__status': 404, 'cookies': None},
                       {'response__status': 200, 'cookies': None},
                       {'response__status': 200, 'cookies': None}])


def test_Collection_QuerySet():
    data = [{'framework': 'Django', 'language': 'Python', 'type': 'full-stack'},
            {'framework': 'Flask', 'language': 'Python', 'type': 'micro'},
            {'framework': 'Rails', 'language': 'Ruby', 'type': 'full-stack'},
            {'framework': 'Sinatra', 'language': 'Ruby', 'type': 'micro'},
            {'framework': 'Zend', 'language': 'PHP', 'type': 'full-stack'},
            {'framework': 'Slim', 'language': 'PHP', 'type': 'micro'}]
    c = Collection(data)
    r1 = c.filter(framework__startswith='S')
    assert isinstance(r1, QuerySet)
    assert len(list(r1)) == 2
    r2 = c.filter(Q(language__exact='Python') | Q(language__exact='Ruby'))
    assert len(list(r2)) == 4
    r3 = c.filter(language='PHP')
    assert_list_equal(list(r3.select('framework', 'type')),
                      [{'framework': 'Zend', 'type': 'full-stack'},
                       {'framework': 'Slim', 'type': 'micro'}])
    r4 = c.filter(Q(language__exact='Python') | Q(language__exact='Ruby'))
    assert_list_equal(list(r4.select('framework')),
                      [{'framework': 'Django'},
                       {'framework': 'Flask'},
                       {'framework': 'Rails'},
                       {'framework': 'Sinatra'}])
    # :todo: test with flatten=True
    r5 = c.filter(framework__startswith='S').select('framework', 'somekey')
    assert_list_equal(list(r5),
                      [{'framework': 'Sinatra', 'somekey': None},
                       {'framework': 'Slim', 'somekey': None}])


## nesdict tests

def test_dunderkey():
    assert dunderkey('a', 'b', 'c') == 'a__b__c'
    assert dunderkey('a') == 'a'
    assert dunderkey('name', 'school_name') == 'name__school_name'


def test_dunder_partition():
    assert dunder_partition('a__b') == ('a', 'b')
    assert dunder_partition('a__b__c') == ('a__b', 'c')
    assert dunder_partition('a') == ('a', None)


def test_dunder_init():
    assert dunder_init('a__b') == 'a'
    assert dunder_init('a__b__c') == 'a__b'
    assert dunder_init('a') == 'a'


def test_dunder_last():
    assert dunder_last('a__b') == 'b'
    assert dunder_last('a__b__c') == 'c'
    assert dunder_last('a') == None


def test_dunder_get():
    d = dict([('a', 'A'),
              ('p', {'q': 'Q'}),
              ('x', {'y': {'z': 'Z'}})])
    assert dunder_get(d, 'a') == 'A'
    assert dunder_get(d, 'p__q') == 'Q'
    assert dunder_get(d, 'x__y__z') == 'Z'


def test_undunder_keys():
    entry = {'request__url': 'http://example.com', 'request__headers': [{'name': 'Connection', 'value': 'Keep-Alive',}],
             'response__status': 404, 'response__headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'}]}
    assert_equal(undunder_keys(entry),
                 {'request': {'url': 'http://example.com', 'headers': [{'name': 'Connection', 'value': 'Keep-Alive',}]},
                  'response': {'status': 404, 'headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'}]}})


def test_dunder_truncate():
    entry = {'request__url': 'http://example.com', 'request__headers': [{'name': 'Connection', 'value': 'Keep-Alive',}],
             'response__status': 404, 'response__headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'}]}
    assert_equal(dunder_truncate(entry),
                 {'url': 'http://example.com',
                  'request__headers': [{'name': 'Connection', 'value': 'Keep-Alive',}],
                  'status': 404,
                  'response__headers': [{'name': 'Date', 'value': 'Thu, 13 Jun 2013 06:43:14 GMT'}]})




"""
   lookupy.lookupy
   ~~~~~~~~~~~~~~~

   This module consists of all of the functionality of the package

"""

import re
from functools import partial

from .dunderkey import dunder_get, dunder_partition, undunder_keys, dunder_truncate


class QuerySet(object):
    """Provides an interface to filter data and select specific fields
    from the data

    QuerySet is used for filtering data and also selecting only
    relevant fields out of it. This object is internally created which
    means usually you, the user wouldn't need to create it.

    :param data: an iterable of dicts

    """

    def __init__(self, data):
        self.data = data

    def filter(self, *args, **kwargs):
        """Filters data using the lookup parameters

        Lookup parameters can be passed as,

          1. keyword arguments of type `field__lookuptype=value` where
             lookuptype specifies how to "query" eg::

                 >>> c.items.filter(language__contains='java')

             above will match all items where the language field
             contains the substring 'java' such as 'java',
             'javascript'. Each look up is treated as a conditional
             clause and if multiple of them are passed, they are
             combined using logical the ``and`` operator

             For nested fields, double underscore can be used eg::

                 >>> data = [{'a': {'b': 3}}, {'a': {'b': 10}}]
                 >>> c = Collection(data)
                 >>> c.items.filter(a__b__gt=5)

             above lookup will match the 2nd element (b > 5)

             For the list of supported lookup parameter, see
             documentation on Github

          2. pos arguments of the type ``field__lookuptype=Q(...)``.
             These can be useful to build conditional clauses that
             need to be combined using logical `or` or negated using
             `not`

                 >>> c.items.filter(Q(language__exact='Python')
                                    |
                                    Q(language__exact='Ruby')

             above query will only filter the data where language is
             either 'Python' or 'Ruby'

        For more documentation see README on Github

        :param args   : ``Q`` objects
        :param kwargs : lookup parameters
        :rtype        : QuerySet

        """
        return self.__class__(filter_items(self.data, *args, **kwargs))

    def select(self, *args, **kwargs):
        """Selects specific fields of the data

        e.g. to select just the keys 'framework' and 'type' from many
        keys, ::

            >>> c.items.select('framework', 'type')


        :param args   : field names to select
        :param kwargs : optional keyword args

        """
        flatten = kwargs.pop('flatten', False)
        f = dunder_truncate if flatten else undunder_keys
        result = (f(d) for d in include_keys(self.data, args))
        return self.__class__(result)

    def __iter__(self):
        for d in self.data:
            yield d


# QuerySet given an alias for backward compatibility
Collection = QuerySet


## filter and lookup functions

def filter_items(items, *args, **kwargs):
    """Filters an iterable using lookup parameters

    :param items  : iterable
    :param args   : ``Q`` objects
    :param kwargs : lookup parameters
    :rtype        : lazy iterable (generator)

    """
    q1 = list(args) if args is not None else []
    q2 = [Q(**kwargs)] if kwargs is not None else []
    lookup_groups = q1 + q2
    pred = lambda item: all(lg.evaluate(item) for lg in lookup_groups)
    return (item for item in items if pred(item))


def lookup(key, val, item):
    """Checks if key-val pair exists in item using various lookup types

    The lookup types are derived from the `key` and then used to check
    if the lookup holds true for the item::

        >>> lookup('request__url__exact', 'http://example.com', item)

    The above will return True if item['request']['url'] ==
    'http://example.com' else False

    :param key  : (str) that represents the field name to find
    :param val  : (mixed) object to match the value in the item against
    :param item : (dict)
    :rtype      : (boolean) True if field-val exists else False

    """
    init, last = dunder_partition(key)
    if last == 'exact':
        return dunder_get(item, init) == val
    elif last == 'neq':
        return dunder_get(item, init) != val
    elif last == 'contains':
        val = guard_str(val)
        return iff_not_none(dunder_get(item, init), lambda y: val in y)
    elif last == 'icontains':
        val = guard_str(val)
        return iff_not_none(dunder_get(item, init), lambda y: val.lower() in y.lower())
    elif last == 'in':
        val = guard_iter(val)
        return dunder_get(item, init) in val
    elif last == 'startswith':
        val = guard_str(val)
        return iff_not_none(dunder_get(item, init), lambda y: y.startswith(val))
    elif last == 'istartswith':
        val = guard_str(val)
        return iff_not_none(dunder_get(item, init), lambda y: y.lower().startswith(val.lower()))
    elif last == 'endswith':
        val = guard_str(val)
        return iff_not_none(dunder_get(item, init), lambda y: y.endswith(val))
    elif last == 'iendswith':
        val = guard_str(val)
        return iff_not_none(dunder_get(item, init), lambda y: y.lower().endswith(val.lower()))
    elif last == 'gt':
        return iff_not_none(dunder_get(item, init), lambda y: y > val)
    elif last == 'gte':
        return iff_not_none(dunder_get(item, init), lambda y: y >= val)
    elif last == 'lt':
        return iff_not_none(dunder_get(item, init), lambda y: y < val)
    elif last == 'lte':
        return iff_not_none(dunder_get(item, init), lambda y: y <= val)
    elif last == 'regex':
        return iff_not_none(dunder_get(item, init), lambda y: re.search(val, y) is not None)
    elif last == 'filter':
        val = guard_Q(val)
        result = guard_list(dunder_get(item, init))
        return len(list(filter_items(result, val))) > 0
    else:
        return dunder_get(item, key) == val


## Classes to compose compound lookups (Q object)

class LookupTreeElem(object):
    """Base class for a child in the lookup expression tree"""

    def __init__(self):
        self.negate = False

    def evaluate(self, item):
        raise NotImplementedError

    def __or__(self, other):
        node = LookupNode()
        node.op = 'or'
        node.add_child(self)
        node.add_child(other)
        return node

    def __and__(self, other):
        node = LookupNode()
        node.add_child(self)
        node.add_child(other)
        return node


class LookupNode(LookupTreeElem):
    """A node (element having children) in the lookup expression tree

    Typically it's any object composed of two ``Q`` objects eg::

        >>> Q(language__neq='Ruby') | Q(framework__startswith='S')
        >>> ~Q(language__exact='PHP')

    """

    def __init__(self):
        super(LookupNode, self).__init__()
        self.children = []
        self.op = 'and'

    def add_child(self, child):
        self.children.append(child)

    def evaluate(self, item):
        """Evaluates the expression represented by the object for the item

        :param item : (dict) item
        :rtype      : (boolean) whether lookup passed or failed

        """
        results = map(lambda x: x.evaluate(item), self.children)
        result = any(results) if self.op == 'or' else all(results)
        return not result if self.negate else result

    def __invert__(self):
        newnode = LookupNode()
        for c in self.children:
            newnode.add_child(c)
        newnode.negate = not self.negate
        return newnode


class LookupLeaf(LookupTreeElem):
    """Class for a leaf in the lookup expression tree"""

    def __init__(self, **kwargs):
        super(LookupLeaf, self).__init__()
        self.lookups = kwargs

    def evaluate(self, item):
        """Evaluates the expression represented by the object for the item

        :param item : (dict) item
        :rtype      : (boolean) whether lookup passed or failed

        """
        result = all(lookup(k, v, item) for k, v in self.lookups.items())
        return not result if self.negate else result

    def __invert__(self):
        newleaf = LookupLeaf(**self.lookups)
        newleaf.negate = not self.negate
        return newleaf


# alias LookupLeaf to Q
Q = LookupLeaf


## functions that work on the keys in a dict

def include_keys(items, fields):
    """Function to keep only specified fields in data

    Returns a list of dict with only the keys mentioned in the
    `fields` param::

        >>> include_keys(items, ['request__url', 'response__status'])

    :param items  : iterable of dicts
    :param fields : (list) fieldnames to keep
    :rtype        : lazy iterable

    """
    return (dict((f, dunder_get(item, f)) for f in fields) for item in items)


## Exceptions

class LookupyError(Exception):
    """Base exception class for all exceptions raised by lookupy"""
    pass


## utility functions

def iff(precond, val, f):
    """If and only if the precond is True

    Shortcut function for precond(val) and f(val). It is mainly used
    to create partial functions for commonly required preconditions

    :param precond : (function) represents the precondition
    :param val     : (mixed) value to which the functions are applied
    :param f       : (function) the actual function

    """
    return False if not precond(val) else f(val)

iff_not_none = partial(iff, lambda x: x is not None)


def guard_type(classinfo, val):
    if not isinstance(val, classinfo):
        raise LookupyError('Value not a {classinfo}'.format(classinfo=classinfo))
    return val

guard_str = partial(guard_type, str)
guard_list = partial(guard_type, list)
guard_Q = partial(guard_type, Q)

def guard_iter(val):
    try:
        iter(val)
    except TypeError:
        raise LookupyError('Value not an iterable')
    else:
        return val


if __name__ == '__main__':
    pass




"""
   lookupy
   ~~~~~~~

   Lookupy is a Python library that provides a Django QuerySet like
   interface to query (select and filter) data (list of dicts)

"""

from .lookupy import Collection, Q

__all__ = ["Collection", "Q"]




from distutils.core import setup

try:
    long_desc = open('./README.rst').read()
except IOError:
    long_desc = 'See: https://github.com/naiquevin/lookupy/blob/master/README.rst'

setup(
    name='Lookupy',
    version='0.1',
    author='Vineet Naik',
    author_email='naikvin@gmail.com',
    url='https://github.com/naiquevin/lookupy',
    packages=['lookupy',],
    license='MIT License',
    description='Django QuerySet inspired interface to query list of dicts',
    long_description=long_desc,
)




## This module deals with code regarding handling the double
## underscore separated keys

def dunderkey(*args):
    """Produces a nested key from multiple args separated by double
    underscore

       >>> dunderkey('a', 'b', 'c')
       >>> 'a__b__c'

    :param *args : *String
    :rtype       : String
    """
    return '__'.join(args)


def dunder_partition(key):
    """Splits a dunderkey into 2 parts

    The first part is everything before the final double underscore
    The second part is after the final double underscore

        >>> dunder_partition('a__b__c')
        >>> ('a__b', 'c')

    :param neskey : String
    :rtype        : 2 Tuple

    """
    parts = key.rsplit('__', 1)
    return tuple(parts) if len(parts) > 1 else (parts[0], None)


def dunder_init(key):
    """Returns the initial part of the dunder key

        >>> dunder_init('a__b__c')
        >>> 'a__b'

    :param neskey : String
    :rtype        : String
    """
    return dunder_partition(key)[0]


def dunder_last(key):
    """Returns the last part of the dunder key

        >>> dunder_last('a__b__c')
        >>> 'c'

    :param neskey : String
    :rtype        : String
    """
    return dunder_partition(key)[1]


def dunder_get(_dict, key):
    """Returns value for a specified dunderkey

    A "dunderkey" is just a fieldname that may or may not contain
    double underscores (dunderscores!) for referrencing nested keys in
    a dict. eg::

         >>> data = {'a': {'b': 1}}
         >>> nesget(data, 'a__b')
         1

    key 'b' can be referrenced as 'a__b'

    :param _dict : (dict)
    :param key   : (str) that represents a first level or nested key in the dict
    :rtype       : (mixed) value corresponding to the key

    """
    parts = key.split('__', 1)
    try:
        result = _dict[parts[0]]
    except KeyError:
        return None
    else:
        return result if len(parts) == 1 else dunder_get(result, parts[1])


def undunder_keys(_dict):
    """Returns dict with the dunder keys converted back to nested dicts

    eg::

        >>> undunder_keys({'a': 'hello', 'b__c': 'world'})
        {'a': 'hello', 'b': {'c': 'world'}}

    :param _dict : (dict) flat dict
    :rtype       : (dict) nested dict

    """
    def f(key, value):
        parts = key.split('__')
        return {
            parts[0]: value if len(parts) == 1 else f(parts[1], value)
        }

    result = {}
    for r in [f(k, v) for k, v in _dict.items()]:
        rk = list(r.keys())[0]
        if rk not in result:
            result.update(r)
        else:
            result[rk].update(r[rk])
    return result


def dunder_truncate(_dict):
    """Returns dict with dunder keys truncated to only the last part

    In other words, replaces the dunder keys with just last part of
    it. In case many identical last parts are encountered, they are
    not truncated further

    eg::

        >>> dunder_truncate({'a__p': 3, 'b__c': 'no'})
        {'c': 'no', 'p': 3}
        >>> dunder_truncate({'a__p': 'yay', 'b__p': 'no', 'c__z': 'dunno'})
        {'a__p': 'yay', 'b__p': 'no', 'z': 'dunno'}

    :param _dict : (dict) to flatten
    :rtype       : (dict) flattened result

    """
    keylist = list(_dict.keys())
    def decide_key(k, klist):
        newkey = dunder_last(k)
        return newkey if list(map(dunder_last, klist)).count(newkey) == 1 else k
    original_keys = [decide_key(key, keylist) for key in keylist]
    return dict(zip(original_keys, _dict.values()))

